@@ -13,32 +13,23 @@
  * See the License for the specific language governing permissions and
  * limitations under the License.
  */
-#include <algorithm>
 #include <fcntl.h>
-#include <ftw.h>
 #include <map>
-#include <stdio.h>
 #include <string>
 #include <sys/stat.h>
-#include <sys/utsname.h>
 #include <unistd.h>
 #include <vector>
+#include <set>
 #include <linux/bpf.h>
+#include <net/if.h>
 
-#include <llvm/ADT/STLExtras.h>
 #include <llvm/ExecutionEngine/MCJIT.h>
 #include <llvm/ExecutionEngine/SectionMemoryManager.h>
-#include <llvm/IRReader/IRReader.h>
-#include <llvm/IR/IRBuilder.h>
 #include <llvm/IR/IRPrintingPasses.h>
 #include <llvm/IR/LegacyPassManager.h>
 #include <llvm/IR/LLVMContext.h>
 #include <llvm/IR/Module.h>
 #include <llvm/IR/Verifier.h>
-#include <llvm/Object/ObjectFile.h>
-#include <llvm/Support/FormattedStream.h>
-#include <llvm/Support/Host.h>
-#include <llvm/Support/SourceMgr.h>
 #include <llvm/Support/TargetSelect.h>
 #include <llvm/Transforms/IPO.h>
 #include <llvm/Transforms/IPO/PassManagerBuilder.h>
@@ -46,14 +37,15 @@
 
 #include "common.h"
 #include "bcc_debug.h"
-#include "bcc_exception.h"
+#include "bcc_elf.h"
 #include "frontends/b/loader.h"
 #include "frontends/clang/loader.h"
 #include "frontends/clang/b_frontend_action.h"
 #include "bpf_module.h"
 #include "exported_files.h"
-#include "kbuild_helper.h"
 #include "libbpf.h"
+#include "bcc_btf.h"
+#include "bcc_libbpf_inc.h"
 
 namespace ebpf {
 
@@ -73,7 +65,7 @@ const string BPFModule::FN_PREFIX = BPF_FN_PREFIX;
 class MyMemoryManager : public SectionMemoryManager {
  public:
 
-  explicit MyMemoryManager(map<string, tuple<uint8_t *, uintptr_t>> *sections)
+  explicit MyMemoryManager(sec_map_def *sections)
       : sections_(sections) {
   }
 
@@ -81,32 +73,40 @@ class MyMemoryManager : public SectionMemoryManager {
   uint8_t *allocateCodeSection(uintptr_t Size, unsigned Alignment,
                                unsigned SectionID,
                                StringRef SectionName) override {
-    uint8_t *Addr = SectionMemoryManager::allocateCodeSection(Size, Alignment, SectionID, SectionName);
-    //printf("allocateCodeSection: %s Addr %p Size %ld Alignment %d SectionID %d\n",
+    // The programs need to change from fake fd to real map fd, so not allocate ReadOnly regions.
+    uint8_t *Addr = SectionMemoryManager::allocateDataSection(Size, Alignment, SectionID, SectionName, false);
+    //printf("allocateDataSection: %s Addr %p Size %ld Alignment %d SectionID %d\n",
     //       SectionName.str().c_str(), (void *)Addr, Size, Alignment, SectionID);
-    (*sections_)[SectionName.str()] = make_tuple(Addr, Size);
+    (*sections_)[SectionName.str()] = make_tuple(Addr, Size, SectionID);
     return Addr;
   }
   uint8_t *allocateDataSection(uintptr_t Size, unsigned Alignment,
                                unsigned SectionID, StringRef SectionName,
                                bool isReadOnly) override {
-    uint8_t *Addr = SectionMemoryManager::allocateDataSection(Size, Alignment, SectionID, SectionName, isReadOnly);
-    //printf("allocateDataSection: %s Addr %p Size %ld Alignment %d SectionID %d RO %d\n",
-    //       SectionName.str().c_str(), (void *)Addr, Size, Alignment, SectionID, isReadOnly);
-    (*sections_)[SectionName.str()] = make_tuple(Addr, Size);
+    // The lines in .BTF.ext line_info, if corresponding to remapped files, will have empty source line.
+    // The line_info will be fixed in place, so not allocate ReadOnly regions.
+    uint8_t *Addr = SectionMemoryManager::allocateDataSection(Size, Alignment, SectionID, SectionName, false);
+    //printf("allocateDataSection: %s Addr %p Size %ld Alignment %d SectionID %d\n",
+    //       SectionName.str().c_str(), (void *)Addr, Size, Alignment, SectionID);
+    (*sections_)[SectionName.str()] = make_tuple(Addr, Size, SectionID);
     return Addr;
   }
-  map<string, tuple<uint8_t *, uintptr_t>> *sections_;
+  sec_map_def *sections_;
 };
 
-BPFModule::BPFModule(unsigned flags, TableStorage *ts)
+BPFModule::BPFModule(unsigned flags, TableStorage *ts, bool rw_engine_enabled,
+                     const std::string &maps_ns, bool allow_rlimit,
+                     const char *dev_name)
     : flags_(flags),
+      rw_engine_enabled_(rw_engine_enabled && bpf_module_rw_engine_enabled()),
       used_b_loader_(false),
+      allow_rlimit_(allow_rlimit),
       ctx_(new LLVMContext),
       id_(std::to_string((uintptr_t)this)),
-      ts_(ts) {
-  InitializeNativeTarget();
-  InitializeNativeTargetAsmPrinter();
+      maps_ns_(maps_ns),
+      ts_(ts), btf_(nullptr) {
+  ifindex_ = dev_name ? if_nametoindex(dev_name) : 0;
+  initialize_rw_engine();
   LLVMInitializeBPFTarget();
   LLVMInitializeBPFTargetMC();
   LLVMInitializeBPFTargetInfo();
@@ -139,335 +139,31 @@ BPFModule::~BPFModule() {
     v->leaf_snprintf = unimplemented_snprintf;
   }
 
+  if (!rw_engine_enabled_) {
+    for (auto section : sections_)
+      delete[] get<0>(section.second);
+  }
+
   engine_.reset();
-  rw_engine_.reset();
+  cleanup_rw_engine();
   ctx_.reset();
   func_src_.reset();
 
+  if (btf_)
+    delete btf_;
+
   ts_->DeletePrefix(Path({id_}));
 }
 
-static void debug_printf(Module *mod, IRBuilder<> &B, const string &fmt, vector<Value *> args) {
-  GlobalVariable *fmt_gvar = B.CreateGlobalString(fmt, "fmt");
-  args.insert(args.begin(), B.CreateInBoundsGEP(fmt_gvar, vector<Value *>({B.getInt64(0), B.getInt64(0)})));
-  args.insert(args.begin(), B.getInt64((uintptr_t)stderr));
-  Function *fprintf_fn = mod->getFunction("fprintf");
-  if (!fprintf_fn) {
-    vector<Type *> fprintf_fn_args({B.getInt64Ty(), B.getInt8PtrTy()});
-    FunctionType *fprintf_fn_type = FunctionType::get(B.getInt32Ty(), fprintf_fn_args, /*isvarArg=*/true);
-    fprintf_fn = Function::Create(fprintf_fn_type, GlobalValue::ExternalLinkage, "fprintf", mod);
-    fprintf_fn->setCallingConv(CallingConv::C);
-    fprintf_fn->addFnAttr(Attribute::NoUnwind);
-  }
-  B.CreateCall(fprintf_fn, args);
-}
-
-static void finish_sscanf(IRBuilder<> &B, vector<Value *> *args, string *fmt,
-                          const map<string, Value *> &locals, bool exact_args) {
-  // fmt += "%n";
-  // int nread = 0;
-  // int n = sscanf(s, fmt, args..., &nread);
-  // if (n < 0) return -1;
-  // s = &s[nread];
-  Value *sptr = locals.at("sptr");
-  Value *nread = locals.at("nread");
-  Function *cur_fn = B.GetInsertBlock()->getParent();
-  Function *sscanf_fn = B.GetInsertBlock()->getModule()->getFunction("sscanf");
-  *fmt += "%n";
-  B.CreateStore(B.getInt32(0), nread);
-  GlobalVariable *fmt_gvar = B.CreateGlobalString(*fmt, "fmt");
-  (*args)[1] = B.CreateInBoundsGEP(fmt_gvar, {B.getInt64(0), B.getInt64(0)});
-  (*args)[0] = B.CreateLoad(sptr);
-  args->push_back(nread);
-  CallInst *call = B.CreateCall(sscanf_fn, *args);
-  call->setTailCall(true);
-
-  BasicBlock *label_true = BasicBlock::Create(B.getContext(), "", cur_fn);
-  BasicBlock *label_false = BasicBlock::Create(B.getContext(), "", cur_fn);
-
-  // exact_args means fail if don't consume exact number of "%" inputs
-  // exact_args is disabled for string parsing (empty case)
-  Value *cond = exact_args ? B.CreateICmpNE(call, B.getInt32(args->size() - 3))
-                           : B.CreateICmpSLT(call, B.getInt32(0));
-  B.CreateCondBr(cond, label_true, label_false);
-
-  B.SetInsertPoint(label_true);
-  B.CreateRet(B.getInt32(-1));
-
-  B.SetInsertPoint(label_false);
-  // s = &s[nread];
-  B.CreateStore(
-      B.CreateInBoundsGEP(B.CreateLoad(sptr), B.CreateLoad(nread, true)), sptr);
-
-  args->resize(2);
-  fmt->clear();
-}
-
-// recursive helper to capture the arguments
-static void parse_type(IRBuilder<> &B, vector<Value *> *args, string *fmt,
-                       Type *type, Value *out,
-                       const map<string, Value *> &locals, bool is_writer) {
-  if (StructType *st = dyn_cast<StructType>(type)) {
-    *fmt += "{ ";
-    unsigned idx = 0;
-    for (auto field : st->elements()) {
-      parse_type(B, args, fmt, field, B.CreateStructGEP(type, out, idx++),
-                 locals, is_writer);
-      *fmt += " ";
-    }
-    *fmt += "}";
-  } else if (ArrayType *at = dyn_cast<ArrayType>(type)) {
-    if (at->getElementType() == B.getInt8Ty()) {
-      // treat i8[] as a char string instead of as an array of u8's
-      if (is_writer) {
-        *fmt += "\"%s\"";
-        args->push_back(out);
-      } else {
-        // When reading strings, scanf doesn't support empty "", so we need to
-        // break this up into multiple scanf calls. To understand it, let's take
-        // an example:
-        // struct Event {
-        //   u32 a;
-        //   struct {
-        //     char x[64];
-        //     int y;
-        //   } b[2];
-        //   u32 c;
-        // };
-        // The writer string would look like:
-        //  "{ 0x%x [ { \"%s\" 0x%x } { \"%s\" 0x%x } ] 0x%x }"
-        // But the reader string needs to restart at each \"\".
-        //  reader0(const char *s, struct Event *val) {
-        //    int nread, rc;
-        //    nread = 0;
-        //    rc = sscanf(s, "{ %i [ { \"%n", &val->a, &nread);
-        //    if (rc != 1) return -1;
-        //    s += nread; nread = 0;
-        //    rc = sscanf(s, "%[^\"]%n", &val->b[0].x, &nread);
-        //    if (rc < 0) return -1;
-        //    s += nread; nread = 0;
-        //    rc = sscanf(s, "\" %i } { \"%n", &val->b[0].y, &nread);
-        //    if (rc != 1) return -1;
-        //    s += nread; nread = 0;
-        //    rc = sscanf(s, "%[^\"]%n", &val->b[1].x, &nread);
-        //    if (rc < 0) return -1;
-        //    s += nread; nread = 0;
-        //    rc = sscanf(s, "\" %i } ] %i }%n", &val->b[1].y, &val->c, &nread);
-        //    if (rc != 2) return -1;
-        //    s += nread; nread = 0;
-        //    return 0;
-        //  }
-        *fmt += "\"";
-        finish_sscanf(B, args, fmt, locals, true);
-
-        *fmt = "%[^\"]";
-        args->push_back(out);
-        finish_sscanf(B, args, fmt, locals, false);
-
-        *fmt = "\"";
-      }
-    } else {
-      *fmt += "[ ";
-      for (size_t i = 0; i < at->getNumElements(); ++i) {
-        parse_type(B, args, fmt, at->getElementType(),
-                   B.CreateStructGEP(type, out, i), locals, is_writer);
-        *fmt += " ";
-      }
-      *fmt += "]";
-    }
-  } else if (isa<PointerType>(type)) {
-    *fmt += "0xl";
-    if (is_writer)
-      *fmt += "x";
-    else
-      *fmt += "i";
-  } else if (IntegerType *it = dyn_cast<IntegerType>(type)) {
-    if (is_writer)
-      *fmt += "0x";
-    if (it->getBitWidth() <= 8)
-      *fmt += "%hh";
-    else if (it->getBitWidth() <= 16)
-      *fmt += "%h";
-    else if (it->getBitWidth() <= 32)
-      *fmt += "%";
-    else
-      *fmt += "%l";
-    if (is_writer)
-      *fmt += "x";
-    else
-      *fmt += "i";
-    args->push_back(is_writer ? B.CreateLoad(out) : out);
-  }
-}
-
-// make_reader generates a dynamic function in the instruction set of the host
-// (not bpf) that is able to convert c-strings in the pretty-print format of
-// make_writer back into binary representations. The encoding of the string
-// takes the llvm ir structure format, which closely maps the c structure but
-// not exactly (no support for unions for instance).
-// The general algorithm is:
-//  pod types (u8..u64)                <= %i
-//  array types
-//   u8[]  no nested quotes :(         <= "..."
-//   !u8[]                             <= [ %i %i ... ]
-//  struct types
-//   struct { u8 a; u64 b; }           <= { %i %i }
-//  nesting is supported
-//   struct { struct { u8 a[]; }; }    <= { "" }
-//   struct { struct { u64 a[]; }; }   <= { [ %i %i .. ] }
-string BPFModule::make_reader(Module *mod, Type *type) {
-  auto fn_it = readers_.find(type);
-  if (fn_it != readers_.end())
-    return fn_it->second;
-
-  // int read(const char *in, Type *out) {
-  //   int n = sscanf(in, "{ %i ... }", &out->field1, ...);
-  //   if (n != num_fields) return -1;
-  //   return 0;
-  // }
-
-  IRBuilder<> B(*ctx_);
-
-  FunctionType *sscanf_fn_type = FunctionType::get(
-      B.getInt32Ty(), {B.getInt8PtrTy(), B.getInt8PtrTy()}, /*isVarArg=*/true);
-  Function *sscanf_fn = mod->getFunction("sscanf");
-  if (!sscanf_fn) {
-    sscanf_fn = Function::Create(sscanf_fn_type, GlobalValue::ExternalLinkage,
-                                 "sscanf", mod);
-    sscanf_fn->setCallingConv(CallingConv::C);
-    sscanf_fn->addFnAttr(Attribute::NoUnwind);
-  }
-
-  string name = "reader" + std::to_string(readers_.size());
-  vector<Type *> fn_args({B.getInt8PtrTy(), PointerType::getUnqual(type)});
-  FunctionType *fn_type = FunctionType::get(B.getInt32Ty(), fn_args, /*isVarArg=*/false);
-  Function *fn =
-      Function::Create(fn_type, GlobalValue::ExternalLinkage, name, mod);
-  auto arg_it = fn->arg_begin();
-  Argument *arg_in = &*arg_it;
-  ++arg_it;
-  arg_in->setName("in");
-  Argument *arg_out = &*arg_it;
-  ++arg_it;
-  arg_out->setName("out");
-
-  BasicBlock *label_entry = BasicBlock::Create(*ctx_, "entry", fn);
-  B.SetInsertPoint(label_entry);
-
-  Value *nread = B.CreateAlloca(B.getInt32Ty());
-  Value *sptr = B.CreateAlloca(B.getInt8PtrTy());
-  map<string, Value *> locals{{"nread", nread}, {"sptr", sptr}};
-  B.CreateStore(arg_in, sptr);
-  vector<Value *> args({nullptr, nullptr});
-  string fmt;
-  parse_type(B, &args, &fmt, type, arg_out, locals, false);
-
-  if (0)
-    debug_printf(mod, B, "%p %p\n", vector<Value *>({arg_in, arg_out}));
-
-  finish_sscanf(B, &args, &fmt, locals, true);
-
-  B.CreateRet(B.getInt32(0));
-
-  readers_[type] = name;
-  return name;
-}
-
-// make_writer generates a dynamic function in the instruction set of the host
-// (not bpf) that is able to pretty-print key/leaf entries as a c-string. The
-// encoding of the string takes the llvm ir structure format, which closely maps
-// the c structure but not exactly (no support for unions for instance).
-// The general algorithm is:
-//  pod types (u8..u64)                => 0x%x
-//  array types
-//   u8[]                              => "..."
-//   !u8[]                             => [ 0x%x 0x%x ... ]
-//  struct types
-//   struct { u8 a; u64 b; }           => { 0x%x 0x%x }
-//  nesting is supported
-//   struct { struct { u8 a[]; }; }    => { "" }
-//   struct { struct { u64 a[]; }; }   => { [ 0x%x 0x%x .. ] }
-string BPFModule::make_writer(Module *mod, Type *type) {
-  auto fn_it = writers_.find(type);
-  if (fn_it != writers_.end())
-    return fn_it->second;
-
-  // int write(int len, char *out, Type *in) {
-  //   return snprintf(out, len, "{ %i ... }", out->field1, ...);
-  // }
-
-  IRBuilder<> B(*ctx_);
-
-  string name = "writer" + std::to_string(writers_.size());
-  vector<Type *> fn_args({B.getInt8PtrTy(), B.getInt64Ty(), PointerType::getUnqual(type)});
-  FunctionType *fn_type = FunctionType::get(B.getInt32Ty(), fn_args, /*isVarArg=*/false);
-  Function *fn =
-      Function::Create(fn_type, GlobalValue::ExternalLinkage, name, mod);
-  auto arg_it = fn->arg_begin();
-  Argument *arg_out = &*arg_it;
-  ++arg_it;
-  arg_out->setName("out");
-  Argument *arg_len = &*arg_it;
-  ++arg_it;
-  arg_len->setName("len");
-  Argument *arg_in = &*arg_it;
-  ++arg_it;
-  arg_in->setName("in");
-
-  BasicBlock *label_entry = BasicBlock::Create(*ctx_, "entry", fn);
-  B.SetInsertPoint(label_entry);
-
-  map<string, Value *> locals{
-      {"nread", B.CreateAlloca(B.getInt64Ty())},
-  };
-  vector<Value *> args({arg_out, B.CreateZExt(arg_len, B.getInt64Ty()), nullptr});
-  string fmt;
-  parse_type(B, &args, &fmt, type, arg_in, locals, true);
-
-  GlobalVariable *fmt_gvar = B.CreateGlobalString(fmt, "fmt");
-
-  args[2] = B.CreateInBoundsGEP(fmt_gvar, vector<Value *>({B.getInt64(0), B.getInt64(0)}));
-
-  if (0)
-    debug_printf(mod, B, "%d %p %p\n", vector<Value *>({arg_len, arg_out, arg_in}));
-
-  vector<Type *> snprintf_fn_args({B.getInt8PtrTy(), B.getInt64Ty(), B.getInt8PtrTy()});
-  FunctionType *snprintf_fn_type = FunctionType::get(B.getInt32Ty(), snprintf_fn_args, /*isVarArg=*/true);
-  Function *snprintf_fn = mod->getFunction("snprintf");
-  if (!snprintf_fn)
-    snprintf_fn = Function::Create(snprintf_fn_type, GlobalValue::ExternalLinkage, "snprintf", mod);
-  snprintf_fn->setCallingConv(CallingConv::C);
-  snprintf_fn->addFnAttr(Attribute::NoUnwind);
-
-  CallInst *call = B.CreateCall(snprintf_fn, args);
-  call->setTailCall(true);
-
-  B.CreateRet(call);
-
-  writers_[type] = name;
-  return name;
-}
-
-unique_ptr<ExecutionEngine> BPFModule::finalize_rw(unique_ptr<Module> m) {
-  Module *mod = &*m;
-
-  run_pass_manager(*mod);
-
-  string err;
-  EngineBuilder builder(move(m));
-  builder.setErrorStr(&err);
-  builder.setUseOrcMCJITReplacement(false);
-  auto engine = unique_ptr<ExecutionEngine>(builder.create());
-  if (!engine)
-    fprintf(stderr, "Could not create ExecutionEngine: %s\n", err.c_str());
-  return engine;
+int BPFModule::free_bcc_memory() {
+  return bcc_free_memory();
 }
 
 // load an entire c file as a module
 int BPFModule::load_cfile(const string &file, bool in_memory, const char *cflags[], int ncflags) {
   ClangLoader clang_loader(&*ctx_, flags_);
   if (clang_loader.parse(&mod_, *ts_, file, in_memory, cflags, ncflags, id_,
-                         *func_src_, mod_src_))
+                         *func_src_, mod_src_, maps_ns_, fake_fd_map_, perf_events_))
     return -1;
   return 0;
 }
@@ -480,78 +176,23 @@ int BPFModule::load_cfile(const string &file, bool in_memory, const char *cflags
 int BPFModule::load_includes(const string &text) {
   ClangLoader clang_loader(&*ctx_, flags_);
   if (clang_loader.parse(&mod_, *ts_, text, true, nullptr, 0, "", *func_src_,
-                         mod_src_))
+                         mod_src_, "", fake_fd_map_, perf_events_))
     return -1;
   return 0;
 }
 
-int BPFModule::annotate() {
+void BPFModule::annotate_light() {
   for (auto fn = mod_->getFunctionList().begin(); fn != mod_->getFunctionList().end(); ++fn)
     if (!fn->hasFnAttribute(Attribute::NoInline))
       fn->addFnAttr(Attribute::AlwaysInline);
 
-  // separate module to hold the reader functions
-  auto m = ebpf::make_unique<Module>("sscanf", *ctx_);
-
   size_t id = 0;
   Path path({id_});
   for (auto it = ts_->lower_bound(path), up = ts_->upper_bound(path); it != up; ++it) {
     TableDesc &table = it->second;
     tables_.push_back(&it->second);
     table_names_[table.name] = id++;
-    GlobalValue *gvar = mod_->getNamedValue(table.name);
-    if (!gvar) continue;
-    if (PointerType *pt = dyn_cast<PointerType>(gvar->getType())) {
-      if (StructType *st = dyn_cast<StructType>(pt->getElementType())) {
-        if (st->getNumElements() < 2) continue;
-        Type *key_type = st->elements()[0];
-        Type *leaf_type = st->elements()[1];
-
-        using std::placeholders::_1;
-        using std::placeholders::_2;
-        using std::placeholders::_3;
-        table.key_sscanf = std::bind(&BPFModule::sscanf, this,
-                                     make_reader(&*m, key_type), _1, _2);
-        table.leaf_sscanf = std::bind(&BPFModule::sscanf, this,
-                                      make_reader(&*m, leaf_type), _1, _2);
-        table.key_snprintf = std::bind(&BPFModule::snprintf, this,
-                                       make_writer(&*m, key_type), _1, _2, _3);
-        table.leaf_snprintf =
-            std::bind(&BPFModule::snprintf, this, make_writer(&*m, leaf_type),
-                      _1, _2, _3);
-      }
-    }
   }
-
-  rw_engine_ = finalize_rw(move(m));
-  if (!rw_engine_)
-    return -1;
-  return 0;
-}
-
-StatusTuple BPFModule::sscanf(string fn_name, const char *str, void *val) {
-  auto fn =
-      (int (*)(const char *, void *))rw_engine_->getFunctionAddress(fn_name);
-  if (!fn)
-    return StatusTuple(-1, "sscanf not available");
-  int rc = fn(str, val);
-  if (rc < 0)
-    return StatusTuple(rc, "error in sscanf: %s", std::strerror(errno));
-  return StatusTuple(rc);
-}
-
-StatusTuple BPFModule::snprintf(string fn_name, char *str, size_t sz,
-                                const void *val) {
-  auto fn = (int (*)(char *, size_t,
-                     const void *))rw_engine_->getFunctionAddress(fn_name);
-  if (!fn)
-    return StatusTuple(-1, "snprintf not available");
-  int rc = fn(str, sz, val);
-  if (rc < 0)
-    return StatusTuple(rc, "error in snprintf: %s", std::strerror(errno));
-  if ((size_t)rc == sz)
-    return StatusTuple(-1, "buffer of size %zd too small", sz);
-  return StatusTuple(0);
 }
 
 void BPFModule::dump_ir(Module &mod) {
@@ -586,16 +227,244 @@ int BPFModule::run_pass_manager(Module &mod) {
   return 0;
 }
 
+void BPFModule::load_btf(sec_map_def &sections) {
+  uint8_t *btf_sec = nullptr, *btf_ext_sec = nullptr;
+  uintptr_t btf_sec_size = 0, btf_ext_sec_size = 0;
+
+  for (auto section: sections) {
+    auto sname = section.first;
+    uint8_t *addr = get<0>(section.second);
+    uintptr_t size = get<1>(section.second);
+
+    if (strcmp(".BTF", sname.c_str()) == 0) {
+      btf_sec = addr;
+      btf_sec_size = size;
+    }
+
+    if (strcmp(".BTF.ext", sname.c_str()) == 0) {
+      btf_ext_sec = addr;
+      btf_ext_sec_size = size;
+    }
+  }
+
+  if (btf_sec == nullptr || btf_ext_sec == nullptr)
+    return;
+
+  // Both .BTF and .BTF.ext ELF sections are present.
+  // The remapped files (the main file and /virtual/include/bcc/helpers.h)
+  // will provide missing source codes in the .BTF.ext line_info table.
+  auto helpers_h = ExportedFiles::headers().find("/virtual/include/bcc/helpers.h");
+  if (helpers_h == ExportedFiles::headers().end()) {
+    fprintf(stderr, "Internal error: missing bcc/helpers.h");
+    return;
+  }
+  std::map<std::string, std::string> remapped_sources;
+  remapped_sources["/virtual/main.c"] = mod_src_;
+  remapped_sources["/virtual/include/bcc/helpers.h"] = helpers_h->second;
+
+  BTF *btf = new BTF(flags_ & DEBUG_BTF, sections);
+  int ret = btf->load(btf_sec, btf_sec_size, btf_ext_sec, btf_ext_sec_size,
+                       remapped_sources);
+  if (ret) {
+    delete btf;
+    return;
+  }
+  btf_ = btf;
+}
+
+int BPFModule::create_maps(std::map<std::string, std::pair<int, int>> &map_tids,
+                           std::map<int, int> &map_fds,
+                           std::map<std::string, int> &inner_map_fds,
+                           bool for_inner_map) {
+  std::set<std::string> inner_maps;
+  if (for_inner_map) {
+    for (auto map : fake_fd_map_) {
+      std::string inner_map_name = get<7>(map.second);
+      if (inner_map_name.size())
+        inner_maps.insert(inner_map_name);
+    }
+  }
+
+  for (auto map : fake_fd_map_) {
+    int fd, fake_fd, map_type, key_size, value_size, max_entries, map_flags;
+    const char *map_name;
+    unsigned int pinned_id;
+    std::string inner_map_name;
+    int inner_map_fd = 0;
+
+    fake_fd     = map.first;
+    map_type    = get<0>(map.second);
+    map_name    = get<1>(map.second).c_str();
+    key_size    = get<2>(map.second);
+    value_size  = get<3>(map.second);
+    max_entries = get<4>(map.second);
+    map_flags   = get<5>(map.second);
+    pinned_id   = get<6>(map.second);
+    inner_map_name = get<7>(map.second);
+
+    if (for_inner_map) {
+      if (inner_maps.find(map_name) == inner_maps.end())
+        continue;
+      if (inner_map_name.size()) {
+        fprintf(stderr, "inner map %s has inner map %s\n",
+                map_name, inner_map_name.c_str());
+        return -1;
+      }
+    } else {
+      if (inner_map_fds.find(map_name) != inner_map_fds.end())
+        continue;
+      if (inner_map_name.size())
+        inner_map_fd = inner_map_fds[inner_map_name];
+    }
+
+    if (pinned_id) {
+        fd = bpf_map_get_fd_by_id(pinned_id);
+    } else {
+        struct bpf_create_map_attr attr = {};
+        attr.map_type = (enum bpf_map_type)map_type;
+        attr.name = map_name;
+        attr.key_size = key_size;
+        attr.value_size = value_size;
+        attr.max_entries = max_entries;
+        attr.map_flags = map_flags;
+        attr.map_ifindex = ifindex_;
+        attr.inner_map_fd = inner_map_fd;
+
+        if (map_tids.find(map_name) != map_tids.end()) {
+          attr.btf_fd = btf_->get_fd();
+          attr.btf_key_type_id = map_tids[map_name].first;
+          attr.btf_value_type_id = map_tids[map_name].second;
+        }
+
+        fd = bcc_create_map_xattr(&attr, allow_rlimit_);
+    }
+
+    if (fd < 0) {
+      fprintf(stderr, "could not open bpf map: %s, error: %s\n",
+              map_name, strerror(errno));
+      return -1;
+    }
+
+    if (for_inner_map)
+      inner_map_fds[map_name] = fd;
+
+    map_fds[fake_fd] = fd;
+  }
+
+  return 0;
+}
+
+int BPFModule::load_maps(sec_map_def &sections) {
+  // find .maps.<table_name> sections and retrieve all map key/value type id's
+  std::map<std::string, std::pair<int, int>> map_tids;
+  if (btf_) {
+    for (auto section : sections) {
+      auto sec_name = section.first;
+      if (strncmp(".maps.", sec_name.c_str(), 6) == 0) {
+        std::string map_name = sec_name.substr(6);
+        unsigned key_tid = 0, value_tid = 0;
+        unsigned expected_ksize = 0, expected_vsize = 0;
+
+        // skip extern maps, which won't be in fake_fd_map_ as they do not
+        // require explicit bpf_create_map.
+        bool is_extern = false;
+        for (auto &t : tables_) {
+          if (t->name == map_name) {
+            is_extern = t->is_extern;
+            break;
+          }
+        }
+        if (is_extern)
+          continue;
+
+        for (auto map : fake_fd_map_) {
+          std::string name;
+
+          name = get<1>(map.second);
+          if (map_name == name) {
+            expected_ksize = get<2>(map.second);
+            expected_vsize = get<3>(map.second);
+            break;
+          }
+        }
+
+        int ret = btf_->get_map_tids(map_name, expected_ksize,
+                                     expected_vsize, &key_tid, &value_tid);
+        if (ret)
+          continue;
+
+        map_tids[map_name] = std::make_pair(key_tid, value_tid);
+      }
+    }
+  }
+
+  // create maps
+  std::map<std::string, int> inner_map_fds;
+  std::map<int, int> map_fds;
+  if (create_maps(map_tids, map_fds, inner_map_fds, true) < 0)
+    return -1;
+  if (create_maps(map_tids, map_fds, inner_map_fds, false) < 0)
+    return -1;
+
+  // update map table fd's
+  for (auto it = ts_->begin(), up = ts_->end(); it != up; ++it) {
+    TableDesc &table = it->second;
+    if (map_fds.find(table.fake_fd) != map_fds.end()) {
+      table.fd = map_fds[table.fake_fd];
+      table.fake_fd = 0;
+    }
+  }
+
+  // update instructions
+  for (auto section : sections) {
+    auto sec_name = section.first;
+    if (strncmp(".bpf.fn.", sec_name.c_str(), 8) == 0) {
+      uint8_t *addr = get<0>(section.second);
+      uintptr_t size = get<1>(section.second);
+      struct bpf_insn *insns = (struct bpf_insn *)addr;
+      int i, num_insns;
+
+      num_insns = size/sizeof(struct bpf_insn);
+      for (i = 0; i < num_insns; i++) {
+        if (insns[i].code == (BPF_LD | BPF_DW | BPF_IMM)) {
+          // change map_fd is it is a ld_pseudo */
+          if (insns[i].src_reg == BPF_PSEUDO_MAP_FD &&
+              map_fds.find(insns[i].imm) != map_fds.end())
+            insns[i].imm = map_fds[insns[i].imm];
+          i++;
+        }
+      }
+    }
+  }
+
+  return 0;
+}
+
 int BPFModule::finalize() {
   Module *mod = &*mod_;
+  sec_map_def tmp_sections,
+      *sections_p;
 
-  mod->setDataLayout("e-m:e-p:64:64-i64:64-n32:64-S128");
   mod->setTargetTriple("bpf-pc-linux");
+#if LLVM_MAJOR_VERSION >= 11
+#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
+  mod->setDataLayout("e-m:e-p:64:64-i64:64-i128:128-n32:64-S128");
+#else
+  mod->setDataLayout("E-m:e-p:64:64-i64:64-i128:128-n32:64-S128");
+#endif
+#else
+#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
+  mod->setDataLayout("e-m:e-p:64:64-i64:64-n32:64-S128");
+#else
+  mod->setDataLayout("E-m:e-p:64:64-i64:64-n32:64-S128");
+#endif
+#endif
+  sections_p = rw_engine_enabled_ ? &sections_ : &tmp_sections;
 
   string err;
   EngineBuilder builder(move(mod_));
   builder.setErrorStr(&err);
-  builder.setMCJITMemoryManager(ebpf::make_unique<MyMemoryManager>(&sections_));
+  builder.setMCJITMemoryManager(ebpf::make_unique<MyMemoryManager>(sections_p));
   builder.setMArch("bpf");
   builder.setUseOrcMCJITReplacement(false);
   engine_ = unique_ptr<ExecutionEngine>(builder.create());
@@ -604,25 +473,51 @@ int BPFModule::finalize() {
     return -1;
   }
 
+#if LLVM_MAJOR_VERSION >= 9
+  engine_->setProcessAllSections(true);
+#else
   if (flags_ & DEBUG_SOURCE)
     engine_->setProcessAllSections(true);
+#endif
 
   if (int rc = run_pass_manager(*mod))
     return rc;
 
   engine_->finalizeObject();
 
+  if (flags_ & DEBUG_SOURCE) {
+    SourceDebugger src_debugger(mod, *sections_p, FN_PREFIX, mod_src_,
+                                src_dbg_fmap_);
+    src_debugger.dump();
+  }
+
+  load_btf(*sections_p);
+  if (load_maps(*sections_p))
+    return -1;
+
+  if (!rw_engine_enabled_) {
+    // Setup sections_ correctly and then free llvm internal memory
+    for (auto section : tmp_sections) {
+      auto fname = section.first;
+      uintptr_t size = get<1>(section.second);
+      uint8_t *tmp_p = NULL;
+      // Only copy data for non-map sections
+      if (strncmp("maps/", section.first.c_str(), 5)) {
+        uint8_t *addr = get<0>(section.second);
+        tmp_p = new uint8_t[size];
+        memcpy(tmp_p, addr, size);
+      }
+      sections_[fname] = make_tuple(tmp_p, size, get<2>(section.second));
+    }
+    engine_.reset();
+    ctx_.reset();
+  }
+
   // give functions an id
   for (auto section : sections_)
     if (!strncmp(FN_PREFIX.c_str(), section.first.c_str(), FN_PREFIX.size()))
       function_names_.push_back(section.first);
 
-  if (flags_ & DEBUG_SOURCE) {
-    SourceDebugger src_debugger(mod, sections_, FN_PREFIX, mod_src_,
-                                src_dbg_fmap_);
-    src_debugger.dump();
-  }
-
   return 0;
 }
 
@@ -764,6 +659,20 @@ unsigned BPFModule::kern_version() const {
 
 size_t BPFModule::num_tables() const { return tables_.size(); }
 
+size_t BPFModule::perf_event_fields(const char *event) const {
+  auto it = perf_events_.find(event);
+  if (it == perf_events_.end())
+    return 0;
+  return it->second.size();
+}
+
+const char * BPFModule::perf_event_field(const char *event, size_t i) const {
+  auto it = perf_events_.find(event);
+  if (it == perf_events_.end() || i >= it->second.size())
+    return nullptr;
+  return it->second[i].c_str();
+}
+
 size_t BPFModule::table_id(const string &name) const {
   auto it = table_names_.find(name);
   if (it == table_names_.end()) return ~0ull;
@@ -935,10 +844,15 @@ int BPFModule::load_b(const string &filename, const string &proto_filename) {
 
   BLoader b_loader(flags_);
   used_b_loader_ = true;
-  if (int rc = b_loader.parse(&*mod_, filename, proto_filename, *ts_, id_))
-    return rc;
-  if (int rc = annotate())
+  if (int rc = b_loader.parse(&*mod_, filename, proto_filename, *ts_, id_,
+                              maps_ns_))
     return rc;
+  if (rw_engine_enabled_) {
+    if (int rc = annotate())
+      return rc;
+  } else {
+    annotate_light();
+  }
   if (int rc = finalize())
     return rc;
   return 0;
@@ -956,8 +870,12 @@ int BPFModule::load_c(const string &filename, const char *cflags[], int ncflags)
   }
   if (int rc = load_cfile(filename, false, cflags, ncflags))
     return rc;
-  if (int rc = annotate())
-    return rc;
+  if (rw_engine_enabled_) {
+    if (int rc = annotate())
+      return rc;
+  } else {
+    annotate_light();
+  }
   if (int rc = finalize())
     return rc;
   return 0;
@@ -971,12 +889,78 @@ int BPFModule::load_string(const string &text, const char *cflags[], int ncflags
   }
   if (int rc = load_cfile(text, true, cflags, ncflags))
     return rc;
-  if (int rc = annotate())
-    return rc;
+  if (rw_engine_enabled_) {
+    if (int rc = annotate())
+      return rc;
+  } else {
+    annotate_light();
+  }
 
   if (int rc = finalize())
     return rc;
   return 0;
 }
 
+int BPFModule::bcc_func_load(int prog_type, const char *name,
+                const struct bpf_insn *insns, int prog_len,
+                const char *license, unsigned kern_version,
+                int log_level, char *log_buf, unsigned log_buf_size,
+                const char *dev_name) {
+  struct bpf_load_program_attr attr = {};
+  unsigned func_info_cnt, line_info_cnt, finfo_rec_size, linfo_rec_size;
+  void *func_info = NULL, *line_info = NULL;
+  int ret;
+
+  attr.prog_type = (enum bpf_prog_type)prog_type;
+  attr.name = name;
+  attr.insns = insns;
+  attr.license = license;
+  if (attr.prog_type != BPF_PROG_TYPE_TRACING &&
+      attr.prog_type != BPF_PROG_TYPE_EXT) {
+    attr.kern_version = kern_version;
+  }
+  attr.log_level = log_level;
+  if (dev_name)
+    attr.prog_ifindex = if_nametoindex(dev_name);
+
+  if (btf_) {
+    int btf_fd = btf_->get_fd();
+    char secname[256];
+
+    ::snprintf(secname, sizeof(secname), ".bpf.fn.%s", name);
+    ret = btf_->get_btf_info(secname, &func_info, &func_info_cnt,
+                             &finfo_rec_size, &line_info,
+                             &line_info_cnt, &linfo_rec_size);
+    if (!ret) {
+      attr.prog_btf_fd = btf_fd;
+      attr.func_info = func_info;
+      attr.func_info_cnt = func_info_cnt;
+      attr.func_info_rec_size = finfo_rec_size;
+      attr.line_info = line_info;
+      attr.line_info_cnt = line_info_cnt;
+      attr.line_info_rec_size = linfo_rec_size;
+    }
+  }
+
+  ret = bcc_prog_load_xattr(&attr, prog_len, log_buf, log_buf_size, allow_rlimit_);
+  if (btf_) {
+    free(func_info);
+    free(line_info);
+  }
+
+  return ret;
+}
+
+int BPFModule::bcc_func_attach(int prog_fd, int attachable_fd,
+                               int attach_type, unsigned int flags) {
+  return bpf_prog_attach(prog_fd, attachable_fd,
+                         (enum bpf_attach_type)attach_type, flags);
+}
+
+int BPFModule::bcc_func_detach(int prog_fd, int attachable_fd,
+                               int attach_type) {
+  return bpf_prog_detach2(prog_fd, attachable_fd,
+                          (enum bpf_attach_type)attach_type);
+}
+
 } // namespace ebpf
