@@ -18,11 +18,10 @@
 from __future__ import print_function
 from bcc import BPF, USDT
 import sys
-import ctypes as ct
 
 # arguments
 def usage():
-    print("USAGE: mysqld_latency PID [min_ms]")
+    print("USAGE: mysqld_qslower PID [min_ms]")
     exit()
 if len(sys.argv) < 2:
     usage()
@@ -82,7 +81,7 @@ int do_done(struct pt_regs *ctx) {
     if (delta >= """ + str(min_ns) + """) {
         // populate and emit data struct
         struct data_t data = {.pid = pid, .ts = sp->ts, .delta = delta};
-        bpf_probe_read(&data.query, sizeof(data.query), (void *)sp->query);
+        bpf_probe_read_user(&data.query, sizeof(data.query), (void *)sp->query);
         events.perf_submit(ctx, &data, sizeof(data));
     }
 
@@ -109,19 +108,11 @@ print("Tracing MySQL server queries for PID %d slower than %s ms..." % (pid,
     min_ms_text))
 print("%-14s %-6s %8s %s" % ("TIME(s)", "PID", "MS", "QUERY"))
 
-class Data(ct.Structure):
-    _fields_ = [
-        ("pid", ct.c_ulonglong),
-        ("ts", ct.c_ulonglong),
-        ("delta", ct.c_ulonglong),
-        ("query", ct.c_char * QUERY_MAX)
-    ]
-
 # process event
 start = 0
 def print_event(cpu, data, size):
     global start
-    event = ct.cast(data, ct.POINTER(Data)).contents
+    event = b["events"].event(data)
     if start == 0:
         start = event.ts
     print("%-14.6f %-6d %8.3f %s" % (float(event.ts - start) / 1000000000,
@@ -130,4 +121,7 @@ def print_event(cpu, data, size):
 # loop with callback to print_event
 b["events"].open_perf_buffer(print_event, page_cnt=64)
 while 1:
-    b.kprobe_poll()
+    try:
+        b.perf_buffer_poll()
+    except KeyboardInterrupt:
+        exit()
