@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/python
 # @lint-avoid-python-3-compatibility-imports
 #
 # vfscount  Count VFS calls ("vfs_*").
@@ -14,7 +14,19 @@
 from __future__ import print_function
 from bcc import BPF
 from time import sleep
+from sys import argv
+def usage():
+    print("USAGE: %s [time]" % argv[0])
+    exit()
 
+interval = 99999999
+if len(argv) > 1:
+    try:
+        interval = int(argv[1])
+        if interval == 0:
+            raise
+    except:  # also catches -h, --help
+        usage()
 # load BPF program
 b = BPF(text="""
 #include <uapi/linux/ptrace.h>
@@ -39,7 +51,7 @@ print("Tracing... Ctrl-C to end.")
 
 # output
 try:
-    sleep(99999999)
+    sleep(interval)
 except KeyboardInterrupt:
     pass
 
